#include "typedef.h"

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <malloc.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "object.h"
#include "edge.h"
#include "field.h"
#include "variable.h"
#include "method.h"
#include "class.h"
#include "directory.h"
#include "source.h"

#if defined(WIN32)
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define close _close
#define alloca _alloca
#define getcwd _getcwd
#else
#include <wchar.h>
#include <errno.h>
#include <unistd.h>
#include <alloca.h>

#define _O_BINARY 0
#endif

extern int		g_dynamic;
extern int		g_fields;
extern int		g_variables;
extern int		g_verbose;
extern char		*g_classpathP;
extern char		*g_java[10];
extern char		**g_envPP;

typedef struct {
	u2T				attribute_name_index;
   	u4T				attribute_length;
   	u1T				*infoP;
} attribute_infoT;

typedef struct {
	u2T				access_flags;
	u2T				name_index;
	u2T				descriptor_index;
	u2T				attributes_count;
	attribute_infoT	*attributesP;
} field_infoT;

typedef field_infoT	method_infoT;

typedef struct {
    cp_infoT  		*m_constant_poolP;
    u2T   	  		*m_interfacesP;
    field_infoT  	*m_fieldsP;
    method_infoT 	*m_methodsP;
	attribute_infoT *m_attributesP;

	u4T			m_magic;
	u2T			m_minor_version;
	u2T			m_major_version;
    u2T			m_constant_pool_count;
    u2T			m_access_flags;
    u2T 		m_this_class;
    u2T			m_super_class;
    u2T			m_interfaces_count;
    u2T			m_fields_count;
    u2T 		m_methods_count;
    u2T			m_attributes_count;
} classFileT;

typedef struct {
	int			start_pc;
	int			line_number;
} lineNumberT;

typedef struct {
	int			start_pc;
	int			end_pc;
	const u2T	*nameP;
	u2T			*descriptorP;
	u2T			index;
} localVariableT;

static u1T	*g_imageP        = 0;
static int	g_image_maxsize  = -1;
static int	g_image_size     = 0;
static u1T	*g_needP         = 0;
static int	g_need_maxsize   = -1;
static int	g_need_size      = 0;

static classFileT	g_cf;

/* Convert the pathP to an absolute path.  If cdP != null then the path
 * is with respect to cwdP/cdP because we did a cd into cdP before
 * building the assember
 */

extern int g_access_used[8];
extern int g_mutators;

static char *
absolute_path(const char *pathP)
{
	static char *cwdP     = 0;
	static char	cwd[2048] = {0};
	static char	full_path[2048];

	char		*P, *P1;
	char		c;

	if (*pathP == '/' || (*pathP && pathP[1] == ':')) {
		/* If the path is an absolute path we are done */
		strcpy(full_path, pathP);
	} else {
		if (!cwdP) {
			cwdP = getcwd(cwd, sizeof(cwd));
			if (!cwdP) {
				fprintf(stderr, "Can't identify cwd\n");
				error_exit(1);
		}	}
		strcpy(full_path, cwdP);
		P = full_path + strlen(cwdP);
		*P++ = '/';
		strcpy(P, pathP);
	}

	for (P = full_path; c = *P; ++P) {
		if (c == '\\') {
			*P = '/';
	}	}

	for (P = full_path; ; ) {
		for (P1 = P; c = *P1 && c != '.'; ++P1);
		if (!c) {
			break;
		}
		switch (P1[1]) {
		case '/':
			if (P1 == full_path || P1[-1] == '/') {
				// Erase ./
				P = P1;
				strcpy(P, P+2);
				continue;
			}
			break;
		case '.':
			if (P1[2] == '/' && P1 > full_path+1 && P1[-1] == '/') {
				// Remove <parent>/../
				for (P = P1-2; P != pathP && P[-1] != '/'; --P);
				strcpy(P, P1+3);
				continue;
			}
			break;
		}
		P = P1 + 1;
	}
	return(full_path);
}

const char *
constantName(int i)
{
	static char	buffer[32];

	static const char *constant_name[] = {
		"? 0 ?",
		"Utf8",
		"? 2 ?",
		"int",
		"float",
		"long",
		"double",
		"class",
		"string",
		"field",
		"method",
		"interface",
		"name&type"
	};
	if (i < 0 || i > 12) {
		sprintf(buffer, "? %d ?", i);
		return(buffer);
	}
	return(constant_name[i]);
}

#define Align(X) ((X) + 3) & ~3

u1T	*
getU2(u1T *startP, u2T *valP)
{
	register	u1T	*P = startP;
	register	u2T	val;

	assert(!(((int) valP) & 1));
	val   = (*P++ << 8);
	val  |= *P++;
	*valP = val;
	return(P);
}
	
u1T	*
getU4(u1T *startP, u4T *valP)
{
	register	u1T	*P = startP;
	register	u4T	val;

	assert(!(((int) valP) & 3));
	val   = (*P++ << 24);
	val  |= (*P++ << 16);
	val  |= (*P++ << 8);
	val  |= *P++;
	*valP = val;
	return (P);
}
	
u1T	*
getU8(u1T *startP, u8T *valP)
{
	register	u1T	*P = startP;

	u8T	val;

//	assert(!(((int) valP) & 7));
	val   = (((u8T) *P) << 56); ++P;
	val  |= (((u8T) *P) << 48); ++P;
	val  |= (((u8T) *P) << 40); ++P;
	val  |= (((u8T) *P) << 32); ++P;
	val  |= (((u8T) *P) << 24); ++P;
	val  |= (((u8T) *P) << 16); ++P;
	val  |= (((u8T) *P) << 8);  ++P;
	val  |= *P++;
	*valP = val;
	return (P);
}

/*
  4.4.7 The CONSTANT_Utf8_info Structure

  The CONSTANT_Utf8_info structure is used to represent constant string values. UTF-8 strings are encoded so that character 
  sequences that contain only non-null ASCII characters can be represented using only 1 byte per character, but characters 
  of up to 16 bits can be represented. All characters in the range '\u0001' to '\u007F' are represented by a single byte: 

0  bits 6-0  

The 7 bits of data in the byte give the value of the character represented. 
The null character ('\u0000') and characters in the range '\u0080' to '\u07FF' are represented by a pair of bytes x and y:

x:  1  1  0  bits 10-6  
y:  1  0  bits 5-0  

The bytes represent the character with the value ((x & 0x1f) << 6) + (y & 0x3f). 

Characters in the range '\u0800' to '\uFFFF' are represented by 3 bytes x, y, and z:

x:  1  1  1  0  bits 15-12  
y:  1  0  bits 11-6  
z:  1  0  bits 5-0  

The character with the value ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f) is represented by the bytes. 

The bytes of multibyte characters are stored in the class file in big-endian (high byte first) order.

*/

int
utf8lth(int	bytes, const u1T *compressedP)
{
	const u1T	*P    = compressedP;
	const u1T	*endP = P + bytes;
	int			ret   = 0;
	u1T			c;

	for (; P < endP; ++P) {
		c = *P;
		if (c & 0x80) {
			if (!(c & 0x20)) {
				// Two   byte encoding
				assert((c & 0xe0) == 0xc0);		// x
				c = *++P;
				assert((c & 0xc0) == 0x80);		// y
			} else {
				// Three byte encoding
				assert((c & 0xf0) == 0xe0);		// x
				c = *++P;
				assert((c & 0xc0) == 0x80);		// y
				c = *++P;
				assert((c & 0xc0) == 0x80);		// z
		}	}
		++ret;
	}
	assert(P == endP);
	return(ret);
}

u1T *
getUtf8(int bytes, u1T *compressedP, u2T **stringPP)
{
	u1T	*P       = compressedP;
	u1T	*endP    = P + bytes;
	u2T *startP  = *stringPP;
	u2T	*stringP = startP;
	u1T	c;
	u2T	val;

	for (; P < endP; ++P) {
		c = *P;
		if (!(c & 0x80)) {
			// One byte encoding
			val = c;
		} else {
			if (!(c & 0x20)) {
				// Two byte encoding
				val  = (c & 0x1F) << 6;
			} else {
				// Three byte encoding
				val  = (c & 0xF) << 12;
				c    = *++P;
				val |= (c & 0x3F) << 6;
			}
			c    = *++P;
			val |= (c & 0x3F);
		}
		*stringP++ = val;
	}
	*stringP++ = 0;
	if ((stringP - startP) & 1) {
		// Align
		*stringP++ = 0;
	}

	*stringPP  = stringP;
	assert(P == endP);
	return(endP);
}

static u2T *
get_Utf8(int index)
{
	cp_infoT	*cpP = g_cf.m_constant_poolP + index;
	u2T			*stringP;

	assert(index > 0 && index < g_cf.m_constant_pool_count);
	assert(cpP->tag == CONSTANT_Utf8);
	stringP = cpP->u.stringP;
	return(stringP);
}

static void
dump_Utf8(int index)
{
	dumpUnicode(get_Utf8(index));
}

static u2T *
get_classname(int index)
{
	cp_infoT	*cpP = g_cf.m_constant_poolP + index;
	u2T			*classnameP;

	assert(index > 0 && index < g_cf.m_constant_pool_count);
	assert(cpP->tag == CONSTANT_Class);
	classnameP = get_Utf8(cpP->u.ref.index1);
	return(classnameP);
}

static Cclass *
get_class(int index)
{
	return Cclass::locate(get_classname(index));
}

static void
process_fieldref(Cmethod *methodP, int index, edgetypeE code, u4T current_lineno, int mutator)
{
	cp_infoT	*cpP = g_cf.m_constant_poolP + index;
	Cclass		*classP;
	u2T			*fieldnameP, *fieldtypeP;
	int			index2;
	Cfield		*fieldP;

	assert(index > 0 && index < g_cf.m_constant_pool_count);
	assert(cpP->tag == CONSTANT_Fieldref);

	classP     = get_class(cpP->u.ref.index1);
	index2     = cpP->u.ref.index2;
	assert(index2 > 0 && index2 < g_cf.m_constant_pool_count);
	cpP        = g_cf.m_constant_poolP + index2;
	assert(cpP->tag == CONSTANT_NameAndType);
	fieldnameP = get_Utf8(cpP->u.ref.index1);
	fieldtypeP = get_Utf8(cpP->u.ref.index2);

	fieldP    = Cfield::locate(classP, fieldnameP, fieldtypeP);			// Reference to a field
	
	if (mutator) {
		Cedge::locate(methodP, fieldP, code, current_lineno);
	} else {
		Cedge::locate(fieldP, methodP, code, current_lineno);
	}
}

static void
process_methodref(Cmethod *methodP, int index, edgetypeE code, u4T current_lineno)
{
	cp_infoT	*cpP = g_cf.m_constant_poolP + index;
	Cclass		*classP;
	u2T			*methodnameP, *methodtypeP;
	int			index2;
	Cmethod		*callsP;

	assert(index > 0 && index < g_cf.m_constant_pool_count);
	assert(cpP->tag == CONSTANT_Methodref);
	classP     = get_class(cpP->u.ref.index1);
	index2     = cpP->u.ref.index2;
	assert(index2 > 0 && index2 < g_cf.m_constant_pool_count);
	cpP        = g_cf.m_constant_poolP + index2;
	assert(cpP->tag == CONSTANT_NameAndType);
	methodnameP = get_Utf8(cpP->u.ref.index1);
	methodtypeP = get_Utf8(cpP->u.ref.index2);

	callsP = Cmethod::locate(classP, methodnameP, methodtypeP);
	Cedge::locate(methodP, callsP, code, current_lineno);
}

static void
process_interfaceref(Cmethod *methodP, int index, edgetypeE code, u4T current_lineno, int nargs)
{
	cp_infoT	*cpP = g_cf.m_constant_poolP + index;
	Cclass		*classP;
	u2T			*interfacenameP, *interfacetypeP;
	int			index2;
	Cmethod		*callsP;

	assert(index > 0 && index < g_cf.m_constant_pool_count);
	assert(cpP->tag == CONSTANT_InterfaceMethodref);
	classP     = get_class(cpP->u.ref.index1);
	index2     = cpP->u.ref.index2;
	assert(index2 > 0 && index2 < g_cf.m_constant_pool_count);
	cpP        = g_cf.m_constant_poolP + index2;
	assert(cpP->tag == CONSTANT_NameAndType);
	interfacenameP = get_Utf8(cpP->u.ref.index1);
	interfacetypeP = get_Utf8(cpP->u.ref.index2);

	callsP = Cmethod::locate(classP, interfacenameP, interfacetypeP);
	Cedge::locate(methodP, callsP, code, current_lineno);
}

static void
process_array(Cmethod *methodP, int index, edgetypeE code, u4T current_lineno, u4T dimensions)
{
	u2T		*classnameP = get_classname(index);
	Cclass	*classP;

	if (*classnameP != '[') {
		classP = Cclass::locate(classnameP, dimensions, 0);
	} else {
		classP = Cclass::locate(classnameP);
	}
	Cedge::locate(methodP, classP, code, current_lineno);
}

static void
process_class(Cmethod *methodP, int index, edgetypeE code, u4T current_lineno)
{
	Cclass		*classP = get_class(index);

	Cedge::locate(methodP, classP, code, current_lineno);
}

static void
process_local_variable(Cclass *classP, Cmethod *methodP, int pc, int variable, localVariableT *localVariablesP, edgetypeE code, u4T current_lineno, int mutator)
{
	localVariableT	*localVariableP;
	const u2T		*nameP;
	u2T				*descriptorP;
	Cvariable		*variableP;

	nameP       = 0;
	descriptorP = 0;
	if (localVariableP = localVariablesP) {
		for (; nameP = localVariableP->nameP; ++localVariableP) {
			if (localVariableP->index == variable && localVariableP->start_pc <= pc && localVariableP->end_pc >= pc) {
				descriptorP = localVariableP->descriptorP;
				break;
	}	}	}
	variableP = Cvariable::locate(methodP, variable, nameP, descriptorP);	// Local variable usage
	if (mutator) {
		Cedge::locate(methodP, variableP, code, current_lineno);
	} else {
		Cedge::locate(variableP, methodP, code, current_lineno);
	}
}

const static u2T	sourceFile[] =
	{'S','o','u','r','c','e','F','i','l','e',0};
const static u2T	constantValue[] =
	{'C','o','n','s','t','a','n','t','V','a','l','u','e',0};
const static u2T	exceptions[] = 
	{'E','x','c','e','p','t','i','o','n','s',0};
const static u2T	lineNumberTable[] =
	{'L','i','n','e','N','u','m','b','e','r','T','a','b','l','e',0};
const static u2T	localVariableTable[] =
	{'L','o','c','a','l','V','a','r','i','a','b','l','e','T','a','b','l','e',0};
const static u2T	innerClasses[] =
	{'I','n','n','e','r','C','l','a','s','s','e','s',0};
const static u2T	enclosingMethod[] = 
	{'E','n','c','l','o','s','i','n','g','M','e','t','h','o','d',0};
const static u2T	code[] =
	{'C','o','d','e',0};

#define U1X       0x01
#define U2X       0x02
#define U4X       0x04
#define WIDEX     0x10
#define BRANCHX   0x20
#define TSWITCHX  0x40
#define LSWITCHX  0x80
#define INVALIDX  0x100
#define MUTATORX  0x200

// Hide the string description - only used in javap2
#define OPCODE(A,B,C,D,E)	{A, B, C, D}
	
const static opcodeT opcodes[] = {

/*  00 (0x00) */ OPCODE("noop",					0, -1, -1, "No-op"),
/*  01 (0x01) */ OPCODE("aconst_null",			0, -1, -1, "Stack ... -> ...,null"),
/*  02 (0x02) */ OPCODE("aconst_m1",			0, -1, -1, "Stack ... -> ...,-1"),
/*  03 (0x03) */ OPCODE("iconst_0",				0, -1, -1, "Stack ... -> ...,[int] 0"),
/*  04 (0x04) */ OPCODE("iconst_1",				0, -1, -1, "Stack ... -> ...,[int] 1"),
/*  05 (0x05) */ OPCODE("iconst_2",				0, -1, -1, "Stack ... -> ...,[int] 2"),
/*  06 (0x06) */ OPCODE("iconst_3",				0, -1, -1, "Stack ... -> ...,[int] 3"),
/*  07 (0x07) */ OPCODE("iconst_4",				0, -1, -1, "Stack ... -> ...,[int] 4"),
/*  08 (0x08) */ OPCODE("iconst_5",				0, -1, -1, "Stack ... -> ...,[int] 5"),
/*  09 (0x09) */ OPCODE("lconst_0",				0, -1, -1, "Stack ... -> ...,[long] 0.H,L"),
/*  10 (0x0a) */ OPCODE("lconst_1",				0, -1, -1, "Stack ... -> ...,[long] 1.H,L"),
/*  11 (0x0b) */ OPCODE("fconst_0",				0, -1, -1, "Stack ... -> ...,[float] 0"),
/*  12 (0x0c) */ OPCODE("fconst_1",				0, -1, -1, "Stack ... -> ...,[float] 1"),
/*  13 (0x0d) */ OPCODE("fconst_2",				0, -1, -1, "Stack ... -> ...,[float] 2"),
/*  14 (0x0e) */ OPCODE("dconst_0",				0, -1, -1, "Stack ... -> ...,[dbl] 0.H,L"),
/*  15 (0x0f) */ OPCODE("dconst_1",				0, -1, -1, "Stack ... -> ...,[dbl] 1.H,L"),
/*  16 (0x10) */ OPCODE("bipush",			  U1X, -1, -1, "Stack ... -> ..., [sbyte] arg"),
/*  17 (0x11) */ OPCODE("sipush",			  U2X, -1, -1, "Stack ... -> ..., [sshort] arg"),
/*  18 (0x12) */ OPCODE("ldc",				  U1X, -1, -1, "Stack ... -> ..., const"),
/*  19 (0x13) */ OPCODE("ldc_w",			  U2X, -1, -1, "Stack ... -> ..., const"),
/*  20 (0x14) */ OPCODE("ldc2_w",			  U2X, -1, -1, "Stack ... -> ..., const.H,L"),
/*  21 (0x15) */ OPCODE("iload",		WIDEX|U1X, -2, 21, "Stack ... -> ..., [int] var"),
/*  22 (0x16) */ OPCODE("lload",		WIDEX|U1X, -2, 22, "Stack ... -> ..., [long] var.H,L"),
/*  23 (0x17) */ OPCODE("fload",		WIDEX|U1X, -2, 23, "Stack ... -> ..., [float] var"),
/*  24 (0x18) */ OPCODE("dload",		WIDEX|U1X, -2, 24, "Stack ... -> ..., [dbl] var.H,L"),
/*  25 (0x19) */ OPCODE("aload",		WIDEX|U1X, -2, 25, "Stack ... -> ..., [ref] var"),
/*  26 (0x1a) */ OPCODE("iload_0",			    0,  0, 21, "Stack ... -> ..., [int] var"),
/*  27 (0x1b) */ OPCODE("iload_1",				0,  1, 21, "Stack ... -> ..., [int] var"),
/*  28 (0x1c) */ OPCODE("iload_2",				0,  2, 21, "Stack ... -> ..., [int] var"),
/*  29 (0x1d) */ OPCODE("iload_3",				0,  3, 21, "Stack ... -> ..., [int] var"),
/*  30 (0x1e) */ OPCODE("lload_0",				0,  0, 22, "Stack ... -> ..., [long] var.H,L"),
/*  31 (0x1f) */ OPCODE("lload_1",				0,  1, 22, "Stack ... -> ..., [long] var.H,L"),
/*  32 (0x20) */ OPCODE("lload_2",				0,  2, 22, "Stack ... -> ..., [long] var.H,L"),
/*  33 (0x21) */ OPCODE("lload_3",				0,  3, 22, "Stack ... -> ..., [long] var.H,L"),
/*  34 (0x22) */ OPCODE("fload_0",				0,  0, 23, "Stack ... -> ..., [float] var"),
/*  35 (0x23) */ OPCODE("fload_1",				0,  1, 23, "Stack ... -> ..., [float] var"),
/*  36 (0x24) */ OPCODE("fload_2",				0,  2, 23, "Stack ... -> ..., [float] var"),
/*  37 (0x25) */ OPCODE("fload_3",				0,  3, 23, "Stack ... -> ..., [float] var"),
/*  38 (0x26) */ OPCODE("dload_0",				0,  0, 24, "Stack ... -> ..., [dbl] var.H,L"),
/*  39 (0x27) */ OPCODE("dload_1",				0,  1, 24, "Stack ... -> ..., [dbl] var.H,L"),
/*  40 (0x28) */ OPCODE("dload_2",				0,  2, 24, "Stack ... -> ..., [dbl] var.H,L"),
/*  41 (0x29) */ OPCODE("dload_3",				0,  3, 24, "Stack ... -> ..., [dbl] var.H,L"),
/*  42 (0x2a) */ OPCODE("aload_0",				0,  0, 25, "Stack ... -> ..., [ref] var"),
/*  43 (0x2b) */ OPCODE("aload_1",				0,  1, 25, "Stack ... -> ..., [ref] var"),
/*  44 (0x2c) */ OPCODE("aload_2",				0,  2, 25, "Stack ... -> ..., [ref] var"),
/*  45 (0x2d) */ OPCODE("aload_3",				0,  3, 25, "Stack ... -> ..., [ref] var"),
/*  46 (0x2e) */ OPCODE("iaload",				0, -1, -1, "Stack ..., a, i -> ..., [int] a[i]"),
/*  47 (0x2f) */ OPCODE("laload",				0, -1, -1, "Stack ..., a, i -> ..., [long] a[i].H,L"),
/*  48 (0x30) */ OPCODE("faload",				0, -1, -1, "Stack ..., a, i -> ..., [float] a[i]"),
/*  49 (0x31) */ OPCODE("daload",				0, -1, -1, "Stack ..., a, i -> ..., [dbl] a[i].H,L"),
/*  50 (0x32) */ OPCODE("aaload",				0, -1, -1, "Stack ..., a, i -> ..., [ref] a[i]"),
/*  51 (0x33) */ OPCODE("baload",				0, -1, -1, "Stack ..., a, i -> ..., [sbyte] a[i]"),
/*  52 (0x34) */ OPCODE("caload",				0, -1, -1, "Stack ..., a, i -> ..., [char] a[i]"),
/*  53 (0x35) */ OPCODE("saload",				0, -1, -1, "Stack ..., a, i -> ..., [sshort] a[i]"),
/*  54 (0x36) */ OPCODE("istore",		WIDEX|U1X|MUTATORX, -2, 54, "Stack ...,int -> ... & var = int"),
/*  55 (0x37) */ OPCODE("lstore",		WIDEX|U1X|MUTATORX, -2, 55, "Stack ...,long.H,L -> ... & var = long"),
/*  56 (0x38) */ OPCODE("fstore",		WIDEX|U1X|MUTATORX, -2, 56, "Stack ...,float -> ... & var = float"),
/*  57 (0x39) */ OPCODE("dstore",		WIDEX|U1X|MUTATORX, -2, 57, "Stack ...,dbl.H,L -> ... & var = dbl"),
/*  58 (0x3a) */ OPCODE("astore",		WIDEX|U1X|MUTATORX, -2, 58, "Stack ...,ref -> ... & var = ref"),
/*  59 (0x3b) */ OPCODE("istore_0",				MUTATORX,  0, 54, "Stack ...,int -> ... & var = int"),
/*  60 (0x3c) */ OPCODE("istore_1",				MUTATORX,  1, 54, "Stack ...,int -> ... & var = int"),
/*  61 (0x3d) */ OPCODE("istore_2",				MUTATORX,  2, 54, "Stack ...,int -> ... & var = int"),
/*  62 (0x3e) */ OPCODE("istore_3",				MUTATORX,  3, 54, "Stack ...,int -> ... & var = int"),
/*  63 (0x3f) */ OPCODE("lstore_0",				MUTATORX,  0, 55, "Stack ...,long.H,L -> ... & var = long"),
/*  64 (0x40) */ OPCODE("lstore_1",				MUTATORX,  1, 55, "Stack ...,long.H,L -> ... & var = long"),
/*  65 (0x41) */ OPCODE("lstore_2",				MUTATORX,  2, 55, "Stack ...,long.H,L -> ... & var = long"),
/*  66 (0x42) */ OPCODE("lstore_3",				MUTATORX,  3, 55, "Stack ...,long.H,L -> ... & var = long"),
/*  67 (0x43) */ OPCODE("fstore_0",				MUTATORX,  0, 56, "Stack ...,float -> ... & var = float"),
/*  68 (0x44) */ OPCODE("fstore_1",				MUTATORX,  1, 56, "Stack ...,float -> ... & var = float"),
/*  69 (0x45) */ OPCODE("fstore_2",				MUTATORX,  2, 56, "Stack ...,float -> ... & var = float"),
/*  70 (0x46) */ OPCODE("fstore_3",				MUTATORX,  3, 56, "Stack ...,float -> ... & var = float"),
/*  71 (0x47) */ OPCODE("dstore_0",				MUTATORX,  0, 57, "Stack ...,dbl.H,L -> ... & var = dbl"),
/*  72 (0x48) */ OPCODE("dstore_1",				MUTATORX,  1, 57, "Stack ...,dbl.H,L -> ... & var = dbl"),
/*  73 (0x49) */ OPCODE("dstore_2",				MUTATORX,  2, 57, "Stack ...,dbl.H,L -> ... & var = dbl"),
/*  74 (0x4a) */ OPCODE("dstore_3",				MUTATORX,  3, 57, "Stack ...,dbl.H,L -> ... & var = dbl"),
/*  75 (0x4b) */ OPCODE("astore_0",				MUTATORX,  0, 58, "Stack ...,ref -> ... & var = ref"),
/*  76 (0x4c) */ OPCODE("astore_1",				MUTATORX,  1, 58, "Stack ...,ref -> ... & var = ref"),
/*  77 (0x4d) */ OPCODE("astore_2",				MUTATORX,  2, 58, "Stack ...,ref -> ... & var = ref"),
/*  78 (0x4e) */ OPCODE("astore_3",				MUTATORX,  3, 58, "Stack ...,ref -> ... & var = ref"),
/*  79 (0x4f) */ OPCODE("iastore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [int] v"),
/*  80 (0x50) */ OPCODE("lastore",				MUTATORX, -1, -1, "Stack ...,a,i, v.H,L -> ... & a[i] = [long] v"),
/*  81 (0x51) */ OPCODE("fastore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [float] v"),
/*  82 (0x52) */ OPCODE("dastore",				MUTATORX, -1, -1, "Stack ...,a,i, v.H,L -> ... & a[i] = [dbl] v"),
/*  83 (0x53) */ OPCODE("aastore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [ref] v"),
/*  84 (0x54) */ OPCODE("bastore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [sbyte] v"),
/*  85 (0x55) */ OPCODE("castore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [char] v"),
/*  86 (0x56) */ OPCODE("sastore",				MUTATORX, -1, -1, "Stack ...,a,i, v -> ... & a[i] = [sshort] v"),
/*  87 (0x57) */ OPCODE("pop",					0, -1, -1, "Stack: ...,x -> ..."),
/*  88 (0x58) */ OPCODE("pop2",					0, -1, -1, "Stack: ...,x2,x1 -> ..."),
/*  89 (0x59) */ OPCODE("dup",					0, -1, -1, "Stack: ...,x -> ...,x, x"),
/*  90 (0x5a) */ OPCODE("dup_x1",				0, -1, -1, "Stack ...,x2,x1 -> ...,x1,x2,x1"),
/*  91 (0x5b) */ OPCODE("dup_x2",				0, -1, -1, "Stack ...,x3,x2,x1 -> ...,x1,x3,x2,x1"),
/*  92 (0x5c) */ OPCODE("dup2",					0, -1, -1, "Stack: ...,x2,x1 -> ...,x2,x1,x2,x1"),
/*  93 (0x5d) */ OPCODE("dup2_x1",				0, -1, -1, "Stack: ...,x3,x2,x1 -> ...,x2,x1,x3,x2,x1"),
/*  94 (0x5e) */ OPCODE("dup2_x2",				0, -1, -1, "Stack ...,x4,x3,x2,x1 -> ...,x2,x1,x4,x3,x2,x1"),
/*  95 (0x5f) */ OPCODE("swap",					0, -1, -1, "Stack ...,x2,x1 -> ...,x1,x2"),
/*  96 (0x60) */ OPCODE("iadd",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2+int1)"),
/*  97 (0x61) */ OPCODE("ladd",					0, -1, -1, "Stack ...,long2.H,L,long1.H.L -> ...,(long2+long1).H,L"),
/*  98 (0x62) */ OPCODE("fadd",					0, -1, -1, "Stack ...,float2,float1 -> ...,(float2+float1)"),
/*  99 (0x63) */ OPCODE("dadd",					0, -1, -1, "Stack ...,dbl2.H,L,dbl1.H.L -> ...,(dbl2+dbl1).H,L"),
/* 100 (0x64) */ OPCODE("isub",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2-int1)"),
/* 101 (0x65) */ OPCODE("lsub",					0, -1, -1, "Stack ...,long2.H,L,long2.H,L -> ...,(long2-long1).H,L"),
/* 102 (0x66) */ OPCODE("fsub",					0, -1, -1, "Stack ...,float2,float1 -> ...,(float2-float1)"			),
/* 103 (0x67) */ OPCODE("dsub",					0, -1, -1, "Stack ...,dbl2.H,L,dbl2.H,L -> ...,(dbl2-dbl1).H,L"),
/* 104 (0x68) */ OPCODE("imul",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2*int1)"),
/* 105 (0x69) */ OPCODE("lmul",					0, -1, -1, "Stack ...,long2.H,L,long1.H.L -> ...,(long2*long1).H,L"),
/* 106 (0x6a) */ OPCODE("fmul",					0, -1, -1, "Stack ...,float2,float1 -> ...,(float2*float1)"),
/* 107 (0x6b) */ OPCODE("dmul",					0, -1, -1, "Stack ...,dbl2.H,L,dbl1.H.L -> ...,(dbl2*dbl1).H,L"),
/* 108 (0x6c) */ OPCODE("idiv",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2/int1)"),
/* 109 (0x6d) */ OPCODE("ldiv",					0, -1, -1, "Stack ...,long2.H,L,long1.H.L -> ...,(long2/long1).H,L"),
/* 110 (0x6e) */ OPCODE("fdiv",					0, -1, -1, "Stack ...,float2,float1 -> ...,(float2/float1)"),
/* 111 (0x6f) */ OPCODE("ddiv",					0, -1, -1, "Stack ...,dbl2.H,L,dbl1.H.L -> ...,(dbl2/dbl1).H,L"),
/* 112 (0x70) */ OPCODE("irem",					0, -1, -1, "Stack ...,int2,int1 -> ...,rem(int2/int1)"),
/* 113 (0x71) */ OPCODE("lrem",					0, -1, -1, "Stack ...,long2.H,L,long1.H.L -> ...,rem(long2/long1).H,L"),
/* 114 (0x72) */ OPCODE("frem",					0, -1, -1, "Stack ...,float2,float1 -> ...,rem(float2/float1)"),
/* 115 (0x73) */ OPCODE("drem",					0, -1, -1, "Stack ...,dbl2.H,L,dbl1.H.L -> ...,rem(dbl2/dbl1).H,L"),
/* 116 (0x74) */ OPCODE("ineg",					0, -1, -1, "Stack ...,int -> ...,(0-int)"),
/* 117 (0x75) */ OPCODE("lneg",					0, -1, -1, "Stack ...,long.H,L -> ...,(0-long).H,L"),
/* 118 (0x76) */ OPCODE("fneg",					0, -1, -1, "Stack ...,float -> ...,(0-float)"),
/* 119 (0x77) */ OPCODE("dneg",					0, -1, -1, "Stack ...,dbl.H,L -> ...,(0-dbl).H,L"),
/* 120 (0x78) */ OPCODE("ishl",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2<<(int1&0x1F))"),
/* 121 (0x79) */ OPCODE("lshl",					0, -1, -1, "Stack ...,long2.H,L,long1/H,L -> ...,(long2<<(long1&0x3F))"),
/* 122 (0x7a) */ OPCODE("ishr",					0, -1, -1, "Stack ...,int2,int1 -> ...,(signed int2>>(int1&0x1F))"),
/* 123 (0x7b) */ OPCODE("lshr",					0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ...,(signed long2>>(long1&0x3F)).H,L"),
/* 124 (0x7c) */ OPCODE("iushr",				0, -1, -1, "Stack ...,int2,int1 -> ...,(unsigned int2>>(int1&0x1F))"),
/* 125 (0x7d) */ OPCODE("lushr",				0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ...,(unsigned long2>>(long1&0x3F)).H,L"),
/* 126 (0x7e) */ OPCODE("iand",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2&int1)"),
/* 127 (0x7f) */ OPCODE("land",					0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ...,(long2&long1).H,L"),
/* 128 (0x80) */ OPCODE("ior",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2|int1)"),
/* 129 (0x81) */ OPCODE("lor",					0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ...,(long2|long1).H,L"),
/* 130 (0x82) */ OPCODE("ixor",					0, -1, -1, "Stack ...,int2,int1 -> ...,(int2^int1)"),
/* 131 (0x83) */ OPCODE("lxor",					0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ...,(long2^long1).H,L"),
/* 132 (0x84) */ OPCODE("iinc",			WIDEX|U1X, -2,132, "var += arg"),
/* 133 (0x85) */ OPCODE("i2l",					0, -1, -1, "Stack ...,int -> ...,long.H,L"),
/* 134 (0x86) */ OPCODE("i2f",					0, -1, -1, "Stack ...,int -> ...,float"),
/* 135 (0x87) */ OPCODE("i2d",					0, -1, -1, "Stack ...,int -> ...,dbl.H,L"),
/* 136 (0x88) */ OPCODE("l2i",					0, -1, -1, "Stack ...,long.H,L -> ...,int"),
/* 137 (0x89) */ OPCODE("l2f",					0, -1, -1, "Stack ...,long.H,L -> ...,float"),
/* 138 (0x8a) */ OPCODE("l2d",					0, -1, -1, "Stack ...,long.H,L -> ...,dbl.H,L"),
/* 139 (0x8b) */ OPCODE("f2i",					0, -1, -1, "Stack ...,float -> ...,int"),
/* 140 (0x8c) */ OPCODE("f2l",					0, -1, -1, "Stack ...,float -> ...,long.H,L"),
/* 141 (0x8d) */ OPCODE("f2d",					0, -1, -1, "Stack ...,float -> ...,dbl.H,L"),
/* 142 (0x8e) */ OPCODE("d2i",					0, -1, -1, "Stack ...,dbl.H,L -> ...,int"),
/* 143 (0x8f) */ OPCODE("d2l",					0, -1, -1, "Stack ...,dbl.H,L -> ...,long.H,L"),
/* 144 (0x90) */ OPCODE("d2f",					0, -1, -1, "Stack ...,dbl.H,L -> ...,float"),
/* 145 (0x91) */ OPCODE("i2b",					0, -1, -1, "Stack ...,int -> ...,(signed (int&0xFF))"),
/* 146 (0x92) */ OPCODE("i2c",					0, -1, -1, "Stack ...,int -> ...,(unsigned (int&0xFFFF))"),
/* 147 (0x93) */ OPCODE("i2s",					0, -1, -1, "Stack ...,int -> ...,(signed (int&0xFFFF))"),
/* 148 (0x94) */ OPCODE("lcmp",					0, -1, -1, "Stack ...,long2.H,L,long1.H,L -> ..., (+1/0/-1 : long2 >|=|< long1)"),
/* 149 (0x95) */ OPCODE("fcmpl",				0, -1, -1, "Stack ...,float2,float1 -> ..., (+1/0/-1=NAN : float2 >|=|< float1)"),
/* 150 (0x96) */ OPCODE("fcmpg",				0, -1, -1, "Stack ...,float2,float1 -> ..., (+1=NAN/0/-1 : float2 >|=|< float1)"),
/* 151 (0x97) */ OPCODE("dcmpl",				0, -1, -1, "Stack ...,dbl2,H,L,dbl1.H,L -> ..., (+1/0/-1=NAN : dbl2 >|=|< dbl1)"),
/* 152 (0x98) */ OPCODE("dcmpg",				0, -1, -1, "Stack ...,dbl2,H,L,dbl1.H,L -> ..., (+1=NAN/0/-1 : dbl2 >|=|< dbl1)"),
/* 153 (0x99) */ OPCODE("ifeq",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int==0"),
/* 154 (0x9a) */ OPCODE("ifne",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int!=0"),
/* 155 (0x9b) */ OPCODE("iflt",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int<0"),
/* 156 (0x9c) */ OPCODE("ifge",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int>=0"),
/* 157 (0x9d) */ OPCODE("ifgt",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int>0"),
/* 158 (0x9e) */ OPCODE("ifle",		  BRANCHX|U2X, -1, -1, "Stack ...,int -> ... & if int<=0"),
/* 159 (0x9f) */ OPCODE("if_icmpeq",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2==int1"),
/* 160 (0xa0) */ OPCODE("if_icmpne",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2!=int1"),
/* 161 (0xa1) */ OPCODE("if_icmplt",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2<int1"),
/* 162 (0xa2) */ OPCODE("if_icmpge",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2>=int1"),
/* 163 (0xa3) */ OPCODE("if_icmpgt",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2>int1"),
/* 164 (0xa4) */ OPCODE("if_icmple",  BRANCHX|U2X, -1, -1, "Stack ...,int2,int1 -> ... & if int2<=int1"),
/* 165 (0xa5) */ OPCODE("if_acmpeq",  BRANCHX|U2X, -1, -1, "Stack ...,ref2,ref1 -> ... & if ref2==ref1"),
/* 166 (0xa6) */ OPCODE("if_acmpne",  BRANCHX|U2X, -1, -1, "Stack ...,ref2,ref1 -> ... & if ref2!=ref1"),
/* 167 (0xa7) */ OPCODE("goto ",	  BRANCHX|U2X, -1, -1, ""),
/* 168 (0xa8) */ OPCODE("jsr",		  BRANCHX|U2X, -2,168, "Stack ... -> ..., returnAddress &"),
/* 169 (0xa9) */ OPCODE("ret",		    WIDEX|U1X, -2,169, "Go to address in var"),
/* 170 (0xaa) */ OPCODE("tableswitch",	 TSWITCHX, -1, -1, "Stack ..., index -> ... & branch"),
/* 171 (0xab) */ OPCODE("lookupswitch",	 LSWITCHX, -1, -1, "Stack ..., key -> ... & branch"),
/* 172 (0xac) */ OPCODE("ireturn",				0, -1, -1, "Stack ...,int -> [empty] & return int to caller"),
/* 173 (0xad) */ OPCODE("lreturn",				0, -1, -1, "Stack ...,long.H,L -> [empty] & return long to caller"),
/* 174 (0xae) */ OPCODE("freturn",				0, -1, -1, "Stack ...,float -> [empty] & return float to caller"),
/* 175 (0xaf) */ OPCODE("dreturn",				0, -1, -1, "Stack ...,dbl.H,L -> [empty] & return dbl to caller"),
/* 176 (0xb0) */ OPCODE("areturn",				0, -1, -1, "Stack ...,ref -> [empty] & return ref to caller"),
/* 177 (0xb1) */ OPCODE("return",				0, -1, -1, "Stack ... -> [empty] & return to caller"),
/* 178 (0xb2) */ OPCODE("getstatic",		  U2X, -1, -1, "Stack ... -> ...,val[H,L] where static Field has val"),
/* 179 (0xb3) */ OPCODE("putstatic",		  U2X|MUTATORX, -1, -1, "Stack ..., val[H,L] -> ... & static Field=val"),
/* 180 (0xb4) */ OPCODE("getfield",			  U2X, -1, -1, "Stack ... -> ...,val[H,L] where Field has val"),
/* 181 (0xb5) */ OPCODE("putfield",			  U2X|MUTATORX, -1, -1, "Stack ..., val[H,L] -> ... & Field=val"),
/* 182 (0xb6) */ OPCODE("invokevirtual",	  U2X, -1, -1, "Stack ...,ref [,arg1 , arg2 ...] -> ... & Function invoked"),
/* 183 (0xb7) */ OPCODE("invokespecial",	  U2X, -1, -1, "Stack ...,ref [,arg1 , arg2 ...] -> ... & Function invoked"),
/* 184 (0xb8) */ OPCODE("invokestatic",		  U2X, -1, -1, "Stack ..., [,arg1 , arg2 ...] -> ... & Function invoked"),
/* 185 (0xb9) */ OPCODE("invokeinterface",    U2X, -1, -1, "Stack ...,ref [,arg1 , arg2 ...] -> ... & Interface invoked"),
/* 186 (0xba) */ OPCODE("xxxunusedxxx1",		0, -1, -1, 0),
/* 187 (0xbb) */ OPCODE("new",				  U2X, -1, -1, "Stack ... -> ...,ref"),
/* 188 (0xbc) */ OPCODE("newarray",			  U1X, -1, -1, "Stack ...,size -> ...,array[size]"),
/* 189 (0xbd) */ OPCODE("anewarray",		  U2X, -1, -1, "Stack ...,size -> ...,array[size]"),
/* 190 (0xbe) */ OPCODE("arraylength",			0, -1, -1, "Stack ...,array -> ...,array.length"),
/* 191 (0xbf) */ OPCODE("athrow",				0, -1, -1, "Stack ...,Throwable ref -> [undefined] & throw"),
/* 192 (0xc0) */ OPCODE("checkcast",		  U2X, -1, -1, "Stack ...,ref -> ...,ref & throw if not castable"),
/* 193 (0xc1) */ OPCODE("instanceof",		  U2X, -1, -1, "Stack ...,ref -> ...,(1 if instanceof else 0)"),
/* 194 (0xc2) */ OPCODE("monitorenter",			0, -1, -1, "Stack ..., ref -> ... & exclusive access to ref"),
/* 195 (0xc3) */ OPCODE("monitorexit",			0, -1, -1, "Stack ..., ref -> ... & release lock on ref"),
/* 196 (0xc4) */ OPCODE("wide",					0, -1, -1, "Treat next instruction as having wider argument"),
/* 197 (0xc5) */ OPCODE("multianewarray",	  U2X, -1, -1, "Stack ...,d1,d2,...dn -> array[d1][d2]...[dn] where n=arg2"),
/* 198 (0xc6) */ OPCODE("ifnull",	  BRANCHX|U2X, -1, -1, "Stack ...,ref -> ... & if ref==null"),
/* 199 (0xc7) */ OPCODE("ifnonnull",  BRANCHX|U2X, -1, -1, "Stack ...,ref -> ... & if ref!=null"),
/* 200 (0xc8) */ OPCODE("goto_w",	  BRANCHX|U4X, -1, -1, ""),
/* 201 (0xc9) */ OPCODE("jsr_w",	  BRANCHX|U4X, -1, -1, "Stack ... -> ..., returnAddress &"),
/* 202 (0xca) */ OPCODE("breakpoint",		    0, -1, -1, 0),	/* Reserved opcodes start here */	
/* 203 (0xcb) */ OPCODE("?203?",		 INVALIDX, -1, -1, 0),
/* 204 (0xcc) */ OPCODE("?204?",		 INVALIDX, -1, -1, 0),
/* 205 (0xcd) */ OPCODE("?205?",		 INVALIDX, -1, -1, 0),
/* 206 (0xce) */ OPCODE("?206?",		 INVALIDX, -1, -1, 0),
/* 207 (0xcf) */ OPCODE("?207?",	     INVALIDX, -1, -1, 0),
/* 208 (0xd0) */ OPCODE("?208?",	     INVALIDX, -1, -1, 0),
/* 209 (0xd1) */ OPCODE("ret_w",	          U2X, -2,209, "Go to return address in variable"),
/* 210 (0xd2) */ OPCODE("?210?",	     INVALIDX, -1, -1, 0),
/* 211 (0xd3) */ OPCODE("?211?",	     INVALIDX, -1, -1, 0),
/* 212 (0xd4) */ OPCODE("?212?",	     INVALIDX, -1, -1, 0),
/* 213 (0xd5) */ OPCODE("?213?",	     INVALIDX, -1, -1, 0),
/* 214 (0xd6) */ OPCODE("?214?",	     INVALIDX, -1, -1, 0),
/* 215 (0xd7) */ OPCODE("?215?",	     INVALIDX, -1, -1, 0),
/* 216 (0xd8) */ OPCODE("?216?",	     INVALIDX, -1, -1, 0),
/* 217 (0xd9) */ OPCODE("?217?",	     INVALIDX, -1, -1, 0),
/* 218 (0xda) */ OPCODE("?218?",	     INVALIDX, -1, -1, 0),
/* 219 (0xdb) */ OPCODE("?219?",	     INVALIDX, -1, -1, 0),
/* 220 (0xdc) */ OPCODE("?220?",	     INVALIDX, -1, -1, 0),
/* 221 (0xdd) */ OPCODE("?221?",	     INVALIDX, -1, -1, 0),
/* 222 (0xde) */ OPCODE("?222?",	     INVALIDX, -1, -1, 0),
/* 223 (0xdf) */ OPCODE("?223?",	     INVALIDX, -1, -1, 0),
/* 224 (0xe0) */ OPCODE("?224?",	     INVALIDX, -1, -1, 0),
/* 225 (0xe1) */ OPCODE("?225?",	     INVALIDX, -1, -1, 0),
/* 226 (0xe2) */ OPCODE("?226?",	     INVALIDX, -1, -1, 0),
/* 227 (0xe3) */ OPCODE("?227?",	     INVALIDX, -1, -1, 0),
/* 228 (0xe4) */ OPCODE("?228?",	     INVALIDX, -1, -1, 0),
/* 229 (0xe5) */ OPCODE("?229?",	     INVALIDX, -1, -1, 0),
/* 230 (0xe6) */ OPCODE("?230?",	     INVALIDX, -1, -1, 0),
/* 231 (0xe7) */ OPCODE("?231?",	     INVALIDX, -1, -1, 0),
/* 232 (0xe8) */ OPCODE("?232?",	     INVALIDX, -1, -1, 0),
/* 233 (0xe9) */ OPCODE("?233?",	     INVALIDX, -1, -1, 0),
/* 234 (0xea) */ OPCODE("?234?",	     INVALIDX, -1, -1, 0),
/* 235 (0xeb) */ OPCODE("?235?",	     INVALIDX, -1, -1, 0),
/* 236 (0xec) */ OPCODE("?236?",	     INVALIDX, -1, -1, 0),
/* 237 (0xed) */ OPCODE("?237?",	     INVALIDX, -1, -1, 0),
/* 238 (0xee) */ OPCODE("?238?",	     INVALIDX, -1, -1, 0),
/* 239 (0xef) */ OPCODE("?239?",	     INVALIDX, -1, -1, 0),
/* 240 (0xf0) */ OPCODE("?240?",	     INVALIDX, -1, -1, 0),
/* 241 (0xf1) */ OPCODE("?241?",	     INVALIDX, -1, -1, 0),
/* 242 (0xf2) */ OPCODE("?242?",	     INVALIDX, -1, -1, 0),
/* 243 (0xf3) */ OPCODE("?243?",	     INVALIDX, -1, -1, 0),
/* 244 (0xf4) */ OPCODE("?244?",	     INVALIDX, -1, -1, 0),
/* 245 (0xf5) */ OPCODE("?245?",	     INVALIDX, -1, -1, 0),
/* 246 (0xf6) */ OPCODE("?246?",	     INVALIDX, -1, -1, 0),
/* 247 (0xf7) */ OPCODE("?247?",	     INVALIDX, -1, -1, 0),
/* 248 (0xf8) */ OPCODE("?248?",	     INVALIDX, -1, -1, 0),
/* 249 (0xf9) */ OPCODE("?249?",	     INVALIDX, -1, -1, 0),
/* 250 (0xfa) */ OPCODE("?250?",	     INVALIDX, -1, -1, 0),
/* 251 (0xfb) */ OPCODE("?251?",	     INVALIDX, -1, -1, 0),
/* 252 (0xfc) */ OPCODE("?252?",	     INVALIDX, -1, -1, 0),
/* 253 (0xfd) */ OPCODE("?253?",	     INVALIDX, -1, -1, 0),
/* 254 (0xfe) */ OPCODE("impdep1",	 	 INVALIDX, -1, -1, 0),
/* 255 (0xff) */ OPCODE("impdep2",	   	 INVALIDX, -1, -1, 0)
};

static void
process_code(Cclass *classP, Cmethod *methodP, u4T length, u1T *infoP)
{
	lineNumberT		*lineNumbersP    = 0;
	localVariableT	*localVariablesP = 0;

	u2T				u2;
	u4T				u4, code_length, branch;
	signed int		s4;
	u1T				u1, *P, *startCodeP, *endCodeP;
	lineNumberT		*lineNumberP;
	u4T				exception_table_length, attributes_count, lineno_changes_at_pc, current_lineno, pc, flags;
	int				code, variable;
	const opcodeT	*opcodeP;
	u4T				wide_index = 0;

	P = getU2(infoP, &methodP->m_max_stack);
	P = getU2(P, &methodP->m_max_locals);

	current_lineno       = 0xFFFFFFFF;
	lineno_changes_at_pc = 0x7FFFFFFF;
	lineNumberP          = 0;
	
	startCodeP = getU4(P, &code_length);
	endCodeP   = P = startCodeP + code_length;

	P = getU2(P, &u2);	
	if (exception_table_length = u2) {
		// TODO: how to capture this information
		unsigned int start_pc, end_pc, handler_pc, i;

		for (i = 0; i < exception_table_length; ++i) {
			P = getU2(P, &u2);
			start_pc = u2;
			P = getU2(P, &u2);
			end_pc = u2;
			P = getU2(P, &u2);
			handler_pc = u2;
			P = getU2(P, &u2);
	}	}

	P = getU2(P, &u2);
	if (attributes_count = u2) {
		u4T				attribute_length;
		const u2T		*nameP;
		u1T				*startP;
		u2T				attribute_name_index;
		unsigned int	i;

		for (i = 0; i < attributes_count; ++i) {
			P      = getU2(P, &attribute_name_index);
			nameP  = get_Utf8(attribute_name_index);
			startP = P = getU4(P, &attribute_length);
			
			if (!unicodecmp(nameP, lineNumberTable)) {
				lineNumberT *endLineNumberP;

				P = getU2(P, &u2);
				lineNumbersP = lineNumberP = (lineNumberT *) alloca(sizeof(lineNumberT) * (u2+1));
				for (endLineNumberP = lineNumberP + u2; lineNumberP < endLineNumberP; ++lineNumberP) {
					P = getU2(P, &u2);
					lineNumberP->start_pc = u2;
					P = getU2(P, &u2);
					lineNumberP->line_number = u2;
				}
				assert ((P - startP) == attribute_length);
				endLineNumberP->line_number = -1;
				endLineNumberP->start_pc    = 0x7FFFFFFF;
				current_lineno = lineNumbersP->line_number;
				if (classP->m_lineno < 0) {
					classP->m_lineno = current_lineno;
				}
				methodP->m_lineno = current_lineno;
				lineNumberP    = lineNumbersP+1;
				lineno_changes_at_pc = lineNumberP->start_pc;
				continue;
			}
			if (!unicodecmp(nameP, localVariableTable)) {
				localVariableT	*localVariableP, *endLocalVariableP;

				P = getU2(P, &u2);
				localVariablesP = localVariableP = (localVariableT *) alloca(sizeof(localVariableT) * (u2 + 1));
				for (endLocalVariableP = localVariableP + u2; localVariableP < endLocalVariableP; ++localVariableP) {
					P = getU2(P, &u2);
					localVariableP->start_pc = u2;
					P = getU2(P, &u2);
					localVariableP->end_pc = localVariableP->start_pc + u2;
					P = getU2(P, &u2);
					localVariableP->nameP = get_Utf8(u2);
					P = getU2(P, &u2);
					localVariableP->descriptorP = get_Utf8(u2);
					P = getU2(P, &localVariableP->index);
				}
				endLocalVariableP->nameP = 0;
				assert ((P - startP) == attribute_length);
				continue;
			}
			P += attribute_length;
	}	}
	
	assert(P - infoP == length);

	for (P = startCodeP; P < endCodeP; ) {
		pc = (u4T) (P - startCodeP);
		if (lineNumberP) {
			while (pc >= lineno_changes_at_pc) {
				current_lineno = lineNumberP->line_number;
				++lineNumberP;
				lineno_changes_at_pc = lineNumberP->start_pc;
		}	}
		code = *P++;
		opcodeP = opcodes + code;
		flags   = opcodeP->flags;

		if (flags & INVALIDX) {
			fprintf(stderr, ": Unknown pcode operation of %d 0x%02x\n", code);
			assert(0);
		}
		if (flags & (U1X|U2X|U4X)) {
			if ((flags & U1X)) {
				if (wide_index) {
					P = getU2(P, &u2);
					u4     = u2;
					s4     = ((signed short) u2);
					if (code != 132) {
						wide_index = 0;
					}
				} else {
					u4     = u1 = *P++;
					s4     = ((signed char) u1);
				}
			} else if (flags & U2X) {
				P = getU2(P, &u2);
				u4     = u2;
				s4     = ((signed short) u2);
			} else {
				P = getU4(P, &u4);
				s4     = (signed int) u4;
			}
			branch = pc + s4;
		}	

		if ((variable = opcodeP->variable) != -1) {
			if (variable < 0) {
				variable = u4;
			}
			if (g_variables) {
				process_local_variable(classP, methodP, P-startCodeP, variable, localVariablesP, (edgetypeE) opcodeP->code1, current_lineno, ((flags&MUTATORX) && g_mutators));
		}	}
		
		// Pushing Constants
		// http://sunsite.ee/java/vmspec/vmspec-14.html#HEADING14-0

		switch (code) {
		case 132:		// iinc
			if (wide_index) {
				P = getU2(P, &u2);
				s4 = ((signed short) u2);
				wide_index = 0;
			} else {
				s4 = *((signed char *) P);
				++P;
			}
			break;
		case 170:	// tableswitch
		{
			s4T	def, low, high, jump;
			s4T	i;

			while ((P-startCodeP) & 3) {
				assert(!*P);
				++P;
			}
			P = getU4(P, (u4T *) &def);	// default
			P = getU4(P, (u4T *) &low);
			P = getU4(P, (u4T *) &high);

			for (i = low; i <= high; ++i) {
				P = getU4(P, (u4T *) &jump);
			}
			break;
		}
		case 171:	// LookupSwitch
		{
			s4T	def, pairs, match, jump, i;

			while ((P - startCodeP) & 3) {
				assert(!*P);
				++P;
			}
			P = getU4(P, (u4T *) &def);		// default
			P = getU4(P, (u4T *) &pairs);	// number of pairs
			for (i = 0; i < pairs; ++i) {
				P = getU4(P, (u4T *) &match);
				P = getU4(P, (u4T *) &jump);
			}
			break;
		}
		case 178:		// getstatic
		case 180:		// getfield
			if (g_fields) {
				process_fieldref(methodP, u4, (edgetypeE) code, current_lineno, 0);
			}
			break;
		case 179:		// putstatic
		case 181:		// putfield
			if (g_fields) {
				process_fieldref(methodP, u4, (edgetypeE) code, current_lineno, g_mutators);
			}
			break;
		case 182:		// Invokevirtual
		case 183:		// Invokespecial
		case 184:		// Invokestatic
			process_methodref(methodP, u4, (edgetypeE) code, current_lineno);
			break;
		case 185:		// Invokeinterface
		{
			int		nargs = *P++;

			process_interfaceref(methodP, u4, (edgetypeE) code, current_lineno, nargs);
			*P++;	// Reserved
			break;
		}

		// Managing arrays
		// http://sunsite.ee/java/vmspec/vmspec-18.html#HEADING18-0

		case 197:		// multianewarray
			u1 = *P++;	// number of dimensions
			process_array(methodP, u4, (edgetypeE) code, current_lineno, u1);
			break;
		case 189:		// anewarray
			process_array(methodP, u4, (edgetypeE) code, current_lineno, 1);
			break;
		case 187:		// new
		case 192:		// checkcast
		case 193:		// instanceof
			process_class(methodP, u4, (edgetypeE) code, current_lineno);
			break;
		case 196:		// wide
			wide_index = 1;
			break;
		}
	}
	assert(P == endCodeP);
}

static void
process_information(const char *filenameP)
{
	cp_infoT		*constant_poolP = g_cf.m_constant_poolP;
	attribute_infoT	*attributesP    = g_cf.m_attributesP;
	attribute_infoT	*end_attributeP = attributesP + g_cf.m_attributes_count;
	field_infoT		*fieldsP        = g_cf.m_fieldsP;
	field_infoT		*end_fieldP     = fieldsP + g_cf.m_fields_count;
	method_infoT	*methodsP       = g_cf.m_methodsP;
	method_infoT	*end_methodP    = methodsP + g_cf.m_methods_count;
	u2T				*sourcenameP    = 0;
	unsigned int	inner_classes   = 0;

	attribute_infoT	*attributeP;
	u2T				*nameP, *classnameP, *outerclassP;
	u2T				*u2P;
	Csource			*sourceP;
	int				count, i;
	Cfield			*fieldP;
	u2T				*descriptorP;
	Cclass			*classP, *class1P;
	Cmethod			*methodP;
	cp_infoT		*cpP;
	u2T				u2, cnt;
	u1T				*P;
	char			*absolute_pathP;
	char			path[2048];
	char			*P1, *P2;
	u2T				access;

	classnameP = get_classname(g_cf.m_this_class);
	classP     = Cclass::locate(classnameP);

	for (attributeP = attributesP; attributeP < end_attributeP; ++attributeP) {
		nameP = get_Utf8(attributeP->attribute_name_index);
		if (!unicodecmp(nameP, sourceFile)) {

			if (sourcenameP) {
				fprintf(stderr, "Multiple source declarations in %s\n", filenameP);
				error_exit(1);
			}
			assert(attributeP->attribute_length == 2);
			getU2(attributeP->infoP, &u2);
			sourcenameP = get_Utf8(u2);
			continue;
		}
		if (!unicodecmp(nameP, innerClasses)) {
			u1T	*P;
			u2T	cnt, inner_class, outer_class, inner_name;

			P = getU2(attributeP->infoP, &cnt);
			for (i = 0; i < cnt; ++i) {
				P = getU2(P, &inner_class);
				P = getU2(P, &outer_class);
				P = getU2(P, &inner_name);
				P = getU2(P, &access);

				if (outer_class) {
					outerclassP = get_classname(outer_class);
					if (!unicodecmp(classnameP, outerclassP)) {
						++inner_classes;
			}	}	}
			assert ((P - attributeP->infoP) == attributeP->attribute_length);
			continue;
		}
		if (!unicodecmp(nameP, enclosingMethod)) {
			u1T		*P;
		    u2T		class_index, method_index;
			Cclass	*outerP;

			P = getU2(attributeP->infoP, &class_index);
			P = getU2(P,                 &method_index);

			assert ((P - attributeP->infoP) == attributeP->attribute_length);

			outerclassP = get_classname(class_index);
			outerP      = Cclass::locate(outerclassP);
			if (method_index) {
				const u2T	*methodnameP, *methodtypeP;

				assert(method_index > 0 && method_index < g_cf.m_constant_pool_count);
				cpP = g_cf.m_constant_poolP + method_index;
				assert(cpP->tag == CONSTANT_NameAndType);
				methodnameP = get_Utf8(cpP->u.ref.index1);
				methodtypeP = get_Utf8(cpP->u.ref.index2);
				classP->set_parent(Cmethod::locate(outerP, methodnameP, methodtypeP));
			} else {
				// The nearest containing thing for this class is the indicated class
				classP->set_parent(outerP);
	}	}	}

	classP->m_processed = 1;

	if (!classP->m_parentP) {
		absolute_pathP = absolute_path(filenameP);
		P1 = strrchr(absolute_pathP, '/');
		if (!P1) {
			P1 = path;
		} else {
			strcpy(path, absolute_pathP);
			P1 = (P1 - absolute_pathP) + path + 1;
		}
		if (sourcenameP) {
			// Replace the class name in filenameP with the source name in nameP
			for (u2P = sourcenameP; *P1++ = (u1T) *u2P++;);
		}

		for (P1 = path;;) {
			switch (*P1) {
			case 0:
				break;
			case '/':
				if (P1[1] == '.') {
					if (!strncmp(P1, "/./", 3)) {
						strcpy(P1, P1+2);
						continue;
					}
					if (!strncmp(P1, "/../", 4)) {
						for (P2 = P1; --P2 >= path && *P2 != '/'; );
						if (P2 >= path) {
							strcpy(P2, P1+3);
							P1 = P2;
							continue;
				}	}	}
			default:
				++P1;
				continue;
			}
			break;
		}

	//	fprintf(stderr,"path=%s\n", path);
		sourceP = Csource::locate(path);
		if (!sourceP) {
			outofmemory();
		}
		classP->set_parent(sourceP);
	}

	classP->m_major_version = g_cf.m_major_version;
	classP->m_minor_version = g_cf.m_minor_version;
	classP->m_access_flags  = g_cf.m_access_flags;

	g_access_used[g_cf.m_access_flags & 7] = 1;

	if (g_cf.m_super_class) {
		classP->m_extendsP = get_class(g_cf.m_super_class);
	}

	if (count = g_cf.m_interfaces_count) {
		u2T	*interfaceP     = g_cf.m_interfacesP;
		u2T	index;

		classP->m_implements_cnt = count;
		classP->m_implementsPP   = (Cclass **) xmalloc(count * sizeof(Cclass *));
		for (i = 0; i < count; ++i) {
			index = interfaceP[i];
			if (index) {
				classP->m_implementsPP[i] = get_class(index);
	}	}	}

	if (inner_classes) {
		innerclassT	*innerclassP;

		classP->m_innerclassesP    = innerclassP = (innerclassT *) xmalloc(sizeof(innerclassT) * inner_classes);
		classP->m_innerclasses_cnt = inner_classes;

		for (attributeP = attributesP; attributeP < end_attributeP; ++attributeP) {
			nameP = get_Utf8(attributeP->attribute_name_index);
			if (!unicodecmp(nameP, innerClasses)) {
				u1T	*P;
				u2T	cnt, inner_class, outer_class, inner_name, access;

				P = getU2(attributeP->infoP, &cnt);
				for (i = 0; i < cnt; ++i) {
					P = getU2(P, &inner_class);
					P = getU2(P, &outer_class);
					P = getU2(P, &inner_name);
					P = getU2(P, &access);

					if (outer_class) {
						outerclassP = get_classname(outer_class);
						if (!unicodecmp(classnameP, outerclassP)) {
							innerclassP->m_innerclassP = class1P = get_class(inner_class);
							innerclassP->m_access      = access;
							class1P->m_access_flags   |= access;
							g_access_used[access & 7] = 1;
							class1P->set_parent(classP);
							++innerclassP;
				}	}	}
				assert((innerclassP - classP->m_innerclassesP) == inner_classes);
				assert ((P - attributeP->infoP) == attributeP->attribute_length);
	}	}	}

	if (g_fields) {

		for (; fieldsP < end_fieldP; ++fieldsP) {
			nameP       = get_Utf8(fieldsP->name_index);
			descriptorP = get_Utf8(fieldsP->descriptor_index);
			fieldP      = Cfield::locate(classP, nameP, descriptorP);	// Declaration of a field
			fieldP->add_created();
			fieldP->m_access_flags = access = fieldsP->access_flags;
			g_access_used[access & 7] = 1;

			fieldP->m_constant.tag = CONSTANT_Null;

			attributesP    = fieldsP->attributesP;
			end_attributeP = attributesP + fieldsP->attributes_count;

			for (attributeP = attributesP; attributeP < end_attributeP; ++attributeP) {
				nameP = get_Utf8(attributeP->attribute_name_index);
				if (!unicodecmp(nameP, constantValue)) {
					assert(attributeP->attribute_length == 2);
					getU2(attributeP->infoP, &u2);
					assert(u2 > 0 && u2 < g_cf.m_constant_pool_count);
					cpP = constant_poolP + u2;
					fieldP->m_constant = *cpP;
					switch (cpP->tag) {
					case CONSTANT_Integer:
					case CONSTANT_Float:
					case CONSTANT_Long:
					case CONSTANT_Double:
						break;
					case CONSTANT_String:
						fieldP->m_constant.u.stringP = unicodedup(get_Utf8(cpP->u.ref.index1));
						break;
					case CONSTANT_Class:
						fieldP->m_constant.u.classP  = get_class(cpP->u.ref.index1);
						break;
					default:
						assert(0);
	}	}	}	}	}

	for (; methodsP < end_methodP; ++methodsP) {
		methodP = Cmethod::locate(classP, get_Utf8(methodsP->name_index), get_Utf8(methodsP->descriptor_index));
		if (!methodP) {
			outofmemory();
		}
		methodP->add_created();
		methodP->m_access_flags = access = methodsP->access_flags;
		g_access_used[access & 7] = 1;

		attributeP = methodsP->attributesP;

		for (end_attributeP = attributeP + methodsP->attributes_count; attributeP < end_attributeP; ++attributeP) {
			nameP = get_Utf8(attributeP->attribute_name_index);
			if (!unicodecmp(nameP, exceptions)) {
				P = getU2(attributeP->infoP, &cnt);
				methodP->m_throws_cnt = cnt;
				methodP->m_throwsPP   = (Cclass **) xmalloc(sizeof(Cclass *) * cnt);
				for (i = 0; i < cnt; ++i) {
					P = getU2(P, &u2);
					methodP->m_throwsPP[i] = get_class(u2);
				}
				assert((P - attributeP->infoP) == attributeP->attribute_length);
				continue;
			}
			if (!unicodecmp(nameP, code)) {
				process_code(classP, methodP, attributeP->attribute_length, attributeP->infoP);
}	}	}	}

/* Read the class image from disk */

static int
read_image(const char *filenameP, struct stat *statsP)
{
	int				fileno;
	int				ret;

	g_image_size = statsP->st_size;

	if (g_image_maxsize < g_image_size) {
		if (g_imageP) {
			free(g_imageP);
		}
		g_imageP = (u1T *) xmalloc(g_image_size + 1);
		g_image_maxsize = g_image_size;
	}
	fileno = open(filenameP, O_RDONLY | _O_BINARY, 0);
	if (fileno == -1) {
		fprintf(stderr, "Can't open %s\n", filenameP);
		goto fail;
	}
	ret = read(fileno, g_imageP, g_image_size);
	if (ret != g_image_size) {
		fprintf(stderr, "Can't read %s\n", filenameP);
		goto fail;
	}

	if (close(fileno)) {
		fprintf(stderr, "Can't close %s\n", filenameP);
		goto fail;
	}

	g_imageP[g_image_size] = 0;
	return(0);
fail:
	return(-1);
}

/* Phase 1: Load basic things and compute space needed for tables */

static int
compute_needed(const char *filenameP)
{
	int				need;

	u1T				*P;
	int				i, j;
	u4T				attribute_length, unicode_length;
	u2T				length, interfaces_count, fields_count, methods_count;
	u2T				attributes_count, u2;

	need         = 0;

	P = getU4(g_imageP, &g_cf.m_magic);

	if (g_cf.m_magic != 0xCAFEBABE) {
		fprintf(stderr, "%s is not a valid class file - has magic value 0x%08x\n", filenameP, g_cf.m_magic);
		goto fail;
	}
	P = getU2(P, &g_cf.m_minor_version);
	P = getU2(P, &g_cf.m_major_version);
	P = getU2(P, &g_cf.m_constant_pool_count);

	g_cf.m_constant_poolP = 0;
	if (g_cf.m_constant_pool_count) {
		need += Align(sizeof(g_cf.m_constant_poolP[0]) * g_cf.m_constant_pool_count);

		for (i = 1; i < g_cf.m_constant_pool_count; ++i) {
//			fprintf(stderr, "%d) %s ", i, constantName(*P));
			switch (*P++) {
			case CONSTANT_Utf8:
				getU2(P, &length);
				P    += 2;
/*
				for (j = 0; j < length; ++j) {
					fputc(P[j], stderr);
				}
*/
				unicode_length = utf8lth(length,P) + 1;
				need += Align(unicode_length * sizeof(u2T));
				P    += length;
				break;
			case CONSTANT_Class:
			case CONSTANT_String:
				getU2(P, &u2);
//					fprintf(stderr, "%u", u2);
				P    += 2;
				break;
			case CONSTANT_Integer:
			case CONSTANT_Float:
			case CONSTANT_NameAndType:
			case CONSTANT_Fieldref:
			case CONSTANT_Methodref:
			case CONSTANT_InterfaceMethodref:
				P    += 4;
				break;
			case CONSTANT_Long:
			case CONSTANT_Double:
				P    += 8;
				// 8 byte values take up two slots in the constant point table
				++i;
				break;
			default:
				fprintf(stderr, "Unknown constant code %d\n", P[-1]);
				goto fail;
			}
//				fputc('\n', stderr);
	}	}
	P = getU2(P, &g_cf.m_access_flags);
	P = getU2(P, &g_cf.m_this_class);
	P = getU2(P, &g_cf.m_super_class);
	P = getU2(P, &g_cf.m_interfaces_count);

	g_cf.m_interfacesP = 0;
	if (interfaces_count = g_cf.m_interfaces_count) {
		need += Align(sizeof(g_cf.m_interfacesP[0]) * interfaces_count);
		P    += sizeof(u2T) * interfaces_count;
	}

	P = getU2(P, &g_cf.m_fields_count);
	g_cf.m_fieldsP = 0;
	if (fields_count = g_cf.m_fields_count) {
		need += Align(sizeof(g_cf.m_fieldsP[0]) * fields_count);

		for (i = 0; i < fields_count; ++i) {
			P    += 6;
			P     = getU2(P, &attributes_count);
			if (attributes_count) {
				need += Align(sizeof(attribute_infoT) * attributes_count);
				for (j = 0; j < attributes_count; ++j) {
					P += 2;	// attribute_name_index
					P  = getU4(P, &attribute_length);
					P += attribute_length;
	}	}	}	}

	P = getU2(P, &g_cf.m_methods_count);
	g_cf.m_methodsP = 0;
	if (methods_count = g_cf.m_methods_count) {
		need += Align(sizeof(g_cf.m_methodsP[0]) * methods_count);
		for (i = 0; i < methods_count; ++i) {
			P    += 6;
			P     = getU2(P, &attributes_count);
			if (attributes_count) {
				need += Align(sizeof(attribute_infoT) * attributes_count);
				for (j = 0; j < attributes_count; ++j) {
					P += 2;	// attribute_name_index
					P  = getU4(P, &attribute_length);
					P += attribute_length;
	}	}	}	}

	P = getU2(P, &g_cf.m_attributes_count);
	g_cf.m_attributesP = 0;
	if (attributes_count = g_cf.m_attributes_count) {
		need += Align(sizeof(attribute_infoT) * attributes_count);
	}

	assert(P - g_imageP <= g_image_size);

	g_need_size = need;
	return(0);
fail:
	return(-1);
}

/* Phase 2: Load tables */

static int
load_tables(void)
{
	u1T				*P, *P1;
	cp_infoT		*cpP;
	field_infoT		*fieldP;
	method_infoT	*methodP;
	u2T				*interfaceP;
	attribute_infoT	*attributeP;
	int				i, j;
	u4T				attribute_length;
	u2T				length, interfaces_count, fields_count, methods_count;
	u2T				attributes_count;

	if (g_need_size > g_need_maxsize) {
		if (g_needP) {
			free(g_needP);
		}
		g_needP = (u1T *) xmalloc(g_need_size);
		g_need_maxsize = g_need_size;
	}
	P  = g_imageP + 10;
	P1 = g_needP;
	if (g_cf.m_constant_pool_count) {
		g_cf.m_constant_poolP = cpP = (cp_infoT *) P1;
		P1 += Align(sizeof(*cpP) * g_cf.m_constant_pool_count);

		cpP->tag = CONSTANT_Null;
		++cpP;
		for (i = 1; i < g_cf.m_constant_pool_count; ++i) {
			cpP->tag = (constantE) *P++;
			switch (cpP->tag) {
			case CONSTANT_Utf8:
				getU2(P, &length);
				P    += 2;
				cpP->u.stringP = (u2T *) P1;

				P     = getUtf8(length, P, (u2T **) &P1);
				break;
			case CONSTANT_Integer:
			case CONSTANT_Float:
				P     = getU4(P, &cpP->u.u4val);
				break;
			case CONSTANT_NameAndType:
				P     = getU2(P, &cpP->u.ref.index1);
				P     = getU2(P, &cpP->u.ref.index2);
				break;
			case CONSTANT_Long:
			case CONSTANT_Double:
				P     = getU8(P, &cpP->u.u8val);
				// 8 byte values take 2 slots in the constant pool table
				++cpP;
				cpP->tag = CONSTANT_Null;
				++i;
				break;
			case CONSTANT_Class:
			case CONSTANT_String:
				P     = getU2(P, &cpP->u.ref.index1);
				break;
			case CONSTANT_Fieldref:
			case CONSTANT_Methodref:
			case CONSTANT_InterfaceMethodref:
				P     = getU2(P, &cpP->u.ref.index1);
				P     = getU2(P, &cpP->u.ref.index2);
				break;
			default:
				fprintf(stderr, "Unknown constant code %d\n", P[-1]);
				goto fail;
			}
			++cpP;
	}	}

	assert(cpP == g_cf.m_constant_poolP + g_cf.m_constant_pool_count);

	P += 8;

	if (interfaces_count = g_cf.m_interfaces_count) {
		interfaceP = (u2T *) P1;
		g_cf.m_interfacesP = interfaceP;
		P1 += Align(sizeof(*interfaceP) * interfaces_count);
		for (i = 0; i < interfaces_count; ++i) {
			P = getU2(P, interfaceP);
			++interfaceP;
	}	}

	P += 2;
	if (fields_count = g_cf.m_fields_count) {
		fieldP = (field_infoT *) P1;
		g_cf.m_fieldsP = fieldP;
		P1 += Align(sizeof(*fieldP) * fields_count);
		for (i = 0; i < fields_count; ++i) {
			P = getU2(P, &fieldP->access_flags);
			P = getU2(P, &fieldP->name_index);
			P = getU2(P, &fieldP->descriptor_index);
			P = getU2(P, &fieldP->attributes_count);
			if (!(attributes_count = fieldP->attributes_count)) {
				fieldP->attributesP = 0;
			} else {
				attributeP = (attribute_infoT *) P1;
				fieldP->attributesP = attributeP;
				P1 += Align(sizeof(*attributeP) * attributes_count);
				for (j = 0; j < attributes_count; ++j) {
					P = getU2(P, &attributeP->attribute_name_index);
					P = getU4(P, &attribute_length);
					attributeP->attribute_length = attribute_length;
					attributeP->infoP = P;
					P += attribute_length;
					++attributeP;
			}	}
			++fieldP;
	}	}

	P += 2;
	if (methods_count = g_cf.m_methods_count) {
		methodP = (method_infoT *) P1;
		g_cf.m_methodsP = methodP;
		P1 += Align(sizeof(*methodP) * methods_count);
		for (i = 0; i < methods_count; ++i) {
			P = getU2(P, &methodP->access_flags);
			P = getU2(P, &methodP->name_index);
			P = getU2(P, &methodP->descriptor_index);
			P = getU2(P, &methodP->attributes_count);
			if (!(attributes_count = methodP->attributes_count)) {
				methodP->attributesP = 0;
			} else {
				attributeP = (attribute_infoT *) P1;
				methodP->attributesP = attributeP;
				P1 += Align(sizeof(*attributeP) * attributes_count);
				for (j = 0; j < attributes_count; ++j) {
					P = getU2(P, &attributeP->attribute_name_index);
					P = getU4(P, &attribute_length);
					attributeP->attribute_length = attribute_length;
					attributeP->infoP = P;
					P += attribute_length;
					++attributeP;
			}	}
			++methodP;
	}	}

	P += 2;
	if (attributes_count = g_cf.m_attributes_count) {
		attributeP = (attribute_infoT *) P1;
		g_cf.m_attributesP = attributeP;
		P1 += Align(sizeof(*attributeP) * attributes_count);
		for (j = 0; j < attributes_count; ++j) {
			P = getU2(P, &attributeP->attribute_name_index);
			P = getU4(P, &attribute_length);
			attributeP->attribute_length = attribute_length;
			attributeP->infoP = P;
			P += attribute_length;
			++attributeP;
	}	}

	assert((P - g_imageP) == g_image_size);
	assert((P1 - g_needP) == g_need_size);
	return(0);
fail:
	return(-1);
}

#ifdef WIN32

#include <process.h>

static int
compile(char *pathP)
{
	char		*sourceP, *P, *P1;
	intptr_t	ret;
	int			i;

	sourceP = pathP;
	P       = strrchr(sourceP, '/');
	P1      = strchr(sourceP, '\\');
	if (P1 > P) {
		P = P1;
	}
	if (!P) {
		P = sourceP;
	}
	P       = strchr(P, '$');
	if (P) {
		int	lth = P-pathP;

		sourceP = (char *) alloca(lth+6);
		strncpy(sourceP, pathP, lth);
		strcpy(sourceP+lth, ".java");
	}	
	g_java[4] = sourceP;

	if (g_verbose) {
		for (i = 0; P = g_java[i]; ++i) {
			fputs(P, stderr);
			fputc(' ', stderr);
		}
		fputc('\n', stderr);
	}

	ret = _spawnvpe(_P_WAIT, g_java[0], g_java+1, g_envPP);
	if (ret) {
		if (!g_verbose) {
			for (i = 0; P = g_java[i]; ++i) {
				fputs(P, stderr);
				fputc(' ', stderr);
		}	}
		fprintf(stderr, "returned %d\n", ret);
	} 
	return(ret);
}
#else
#include <sys/wait.h>

static int
compile(char *pathP)
{
	char		*sourceP, *P;
	int			ret, ret1, pid, status;
	int			i;

	sourceP = pathP;
	P       = strrchr(sourceP, '/');
	if (!P) {
		P = sourceP;
	}
	P       = strchr(P, '$');
	if (P) {
		int	lth = P-pathP;

		sourceP = (char *) alloca(lth+6);
		strncpy(sourceP, pathP, lth);
		strcpy(sourceP+lth, ".java");
	}	
	g_java[4] = sourceP;

	if (g_verbose) {
		for (i = 0; P = g_java[i]; ++i) {
			fputs(P, stderr);
			fputc(' ', stderr);
		}
		fputc('\n', stderr);
	}

	ret  = fork();
	ret1 = -2;

	switch (ret) {
	case 0:
		execvp(g_java[0], g_java);
		fprintf(stderr,"Return from execve %s\n", g_java[0]);
		error_exit(1);
	case -1:
		fprintf(stderr,"Fork failed\n");
		break;
	default:
		pid = wait(&status);
		if (pid != ret) {
			fprintf(stderr,"Wait for %s to finish failed\n", g_java[0]);
		} else if (!WIFEXITED(status)) {
			fprintf(stderr,"%s did not terminate normally\n", g_java[0]);
		} else {
			ret1 = WEXITSTATUS(status);
	}	}
	if (ret1) {
		if (!g_verbose) {
			for (i = 0; P = g_java[i]; ++i) {
				fputs(P, stderr);
				fputc(' ', stderr);
		}	}
		fprintf(stderr, "returned %d\n", ret1);
	} 
	return(ret1);
}
#endif

int
readclassfile(char *pathP, int silent)
{
	struct	stat 	stats;
	char			*bufferP;
	int				lth;
	int				ret;

	if (stat(pathP, &stats)) {
		if (silent) {
			return(1);
		}
		fprintf(stderr, "Can't stat %s\n", pathP);
		return(-1);
	}

	lth = strlen(pathP);
	if (lth > 5 && !strcmp(pathP+lth-5,".java")) {
		if (!g_java[0]) {
			fprintf(stderr,"Use -j option if you want to compile %s\n", pathP);
			return(-1);
		}
		fputs("// ", stdout);
		fputs("Compiled ", stdout);
		fputs(pathP, stdout);
		ret = compile(pathP);
		if (ret) {
			fprintf(stdout, " failed ret=%d", ret);
		}
		fputc('\n', stdout);
		
		bufferP = (char *) alloca(lth+2);
		strcpy(bufferP, pathP);
		strcpy(bufferP+lth-4,"class");
		pathP = bufferP;

		if (stat(pathP, &stats)) {
			if (silent) {
				return(1);
			}
			fprintf(stderr, "Can't stat %s\n", pathP);
			return(-1);
		}
	}

	ret = read_image(pathP, &stats);
	if (ret != 0) {
		goto done;
	}

	fputs("// ", stdout);
	if (silent) {
		fputs("Dynamic ", stdout);
	}
	fputs(pathP, stdout);
	fputc('\n', stdout);

	ret = compute_needed(pathP);
	if (ret != 0) {
		goto done;
	}
	ret = load_tables();
	if (ret != 0) {
		goto done;
	}
	process_information(pathP);
	if (g_verbose) {
		if (silent) {
			fputs("Dynamically processed", stderr);
		} else {
			fputs("Processed", stderr);
		}
		fprintf(stderr, " \"%s\"\n", pathP);
	}
done:
	return(ret);
}

void
pseudo_source(void)
{
	int			i;
	Cclass		*classP;
	char		*classpathP;
	Cdirectory0	*directory0P;
	int			repeat;
	char		path[2048];
	char		*pathP, *P;
	u2T			*P2;
	char		c;

	classpathP = g_classpathP;
	if (!classpathP) {
		classpathP = ".";
	}
	do {
		repeat = 0;
		for (i = 0; i < HASHSIZE; ++i) {
			for (classP = Cclass::g_hash[i]; classP; classP = classP->m_hashP) {
				if (!classP->m_processed) {
					if (g_dynamic) {
						pathP = path;
						for (P = classpathP;; ++P) {
							switch (c = *P) {
							case 0:
							case ';':
							case ':':
							case ',':
								*pathP++ = '/';
								break;
							default:
								*pathP++ = c;
								continue;
							}
							for (P2 = classP->m_nameP; *pathP++ = (u1T) *P2++; );

							--pathP;
							if (g_java[0]) {
								strcpy(pathP, ".java");
							} else {
								strcpy(pathP, ".class");
							}
							switch (readclassfile(path, 1)) {
							case 0:
								repeat = 1;
								break;
							case 1:
								if (c && P[1]) {
									// Wrong path
									pathP = path;
									continue;
								}
								break;
							}
							break;
					}	}
					classP->m_processed = 1;
				}
				if (!classP->m_parentP) {
					directory0P = Cdirectory0::locateForClass(Cclass::g_hash[i], classP->m_nameP);
					classP->set_parent(directory0P);
		}	}	}
	} while (repeat);
}


