#include "typedef.h"
#include <assert.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"

#ifdef TRACK_DOWN_MALLOC_PROBLEM
#include <crtdbg.h>
#endif

void
error_exit(int code)
{
	fprintf(stderr,"Exit code %d\n", code);
	exit(code);
}

void
outofmemory(void)
{
	fprintf(stderr, "Out of memory\n");
	error_exit(4);
}

void *
xmalloc(size_t size)
{
#ifdef TRACK_DOWN_MALLOC_PROBLEM
	static int		initialised_crt = 0;

	if (!initialised_crt) {
		printf("Performing exhaustive checking of memory heap\n");
		initialised_crt = 1;
		_CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
		_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
		_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );
		// Get current flag
		int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );

		// Turn on leak-checking bit
		tmpFlag |= _CRTDBG_LEAK_CHECK_DF;
	//	tmpFlag |= _CRTDBG_CHECK_CRT_DF;
		tmpFlag |= _CRTDBG_CHECK_ALWAYS_DF;

		// Set flag to the new value
		_CrtSetDbgFlag( tmpFlag );
	}
	if (!_CrtCheckMemory()) {
		assert(0);
	}
#endif

	void *retP = malloc(size);

	if (!retP) {
		fprintf(stderr, "Can't malloc %u bytes\n", size);
		outofmemory();
	}
	return retP;
}

void
xfree(void *P)
{
	free(P);
}

void
dumpUnicode(const u2T *stringP)
{
	const register u2T *P;
	u2T					c;
//	wchar_t				c1;

	// Under g++ sizeof(wchar_t) == 4 : YUK
	// Worse fputwc returns an error about being unable to seek when attempting
	// to write a unicode character.  For the present fall back on simply
	// emitting ascii characters, hoping that this is ok.
	// N.B. This is not a problem under windows

	for (P = stringP; ; ++P) {
		switch (c = *P) {
		case 0:
			return;
		case '\n':
			c = 'n';
			break;
		case '\t':
			c = 't';
			break;
		case '\f':
			c = 'f';
			break;
		case '\r':
			c = 'r';
			break;
		case 27:	/* Esc */
			c = 'e';
			break;
		case 127:	/* Delete */
			c = 'd';
			break;
		case '\\':
		case '"':
			break;
		default:
			if (c > 255) {
				fprintf(stdout, "\\u%04x", c);
			} else {
				fputc(c, stdout);
			}
			continue;
		}
		fputc('\\', stdout);
		fputc(c, stdout);
/*
		c1 = c;
		if (c1 != fputwc(c1, stdout)) {
			perror("writing unicode");
		}
 */
	}
}

u4T
unicodelth(const u2T *stringP)
{
	register	const u2T	*P  = stringP;
	register	u4T			lth = 0;

	for (;*P++; ++lth);
	return lth;
}

u2T *
unicodecpy1(u2T *string1P, const char *string2P)
{
	register		  u2T	*P1 = string1P;
	register	const char	*P2 = string2P;

	while (*P1++ = *P2++);
	return P1;
}

u2T *
unicodedup1(const char *stringP)
{
	register	const char	*P = stringP;
	register	u2T			*retP;
	size_t		lth;

	lth = strlen(stringP);
	retP = (u2T *) xmalloc((lth+1) * sizeof(u2T));
	if (!retP) {
		outofmemory();
	}
	unicodecpy1(retP, stringP);
	return(retP);
}

u2T *
unicodedup(const u2T *stringP)
{
	register u2T *retP;
	int			 lth;

	lth = unicodelth(stringP);
	retP = (u2T *) xmalloc((lth+1) * sizeof(u2T));
	unicodecpy(retP, stringP);
	return(retP);
}

u2T *
unicodecpy(u2T *string1P, const u2T *string2P)
{
	register		  u2T	*P1 = string1P;
	register	const u2T	*P2 = string2P;

	while (*P1++ = *P2++);
	return P1;
}

int 
unicodecmp(const u2T *string1P, const u2T *string2P)
{
	register	const u2T	*P1 = string1P;
	register	const u2T	*P2 = string2P;

	for (;;) {
		if (!*P1) {
			if (!*P2) {
				return(0);
			}
			return(1);
		}
		if (!*P1) {
			return(1);
		}
		if (*P1 != *P2) {
			return(1);
		}
		++P1;
		++P2;
}	}

void
dump_access(u2T	access)
{
	static const char *accessName[] = {
		/* 0x0001 */	"public",			// ACC_PUBLIC 0x0001 Visible to everyone  Class, Method, Variable 
		/* 0x0002 */	"private",			// ACC_PRIVATE 0x0002 Visible only to the defining class  Method, Variable 
		/* 0x0004 */	"protected",		// ACC_PROTECTED 0x0004 Visible to subclasses  Method, Variable 
		/* 0x0008 */	"static",			// ACC_STATIC 0x0008 Variable or method is static  Method, Variable
		/* 0x0010 */	"final",			// ACC_FINAL 0x0010 No further subclassing, overriding, or assignment after initialization Class, Method, Variable 
		/* 0x0020 */	"synchronized",		// ACC_SYNCHRONIZED 0x0020 Wrap use in monitor lock Method 
		/* 0x0040 */	"volatile",			// ACC_VOLATILE 0x0040 Can't cache Variable 
		/* 0x0080 */	"transient",		// ACC_TRANSIENT 0x0080 Not to be written or read by a persistent object manager  Variable
		/* 0x0100 */	"native",			// ACC_NATIVE 0x0100 Implemented in a language other than Java Method 
		/* 0x0200 */	"interface",		// ACC_INTERFACE 0x0200 Is an interface  Class 
		/* 0x0400 */	"abstract",			// ACC_ABSTRACT 0x0400 No body provided  Class, Method 
		/* 0x0800 */	"strict",			// ACC_STRICT 0x0800 Declared strictfp; floating-point mode is FP-strict 
		/* 0x1000 */	"synthetic",		// ACC_SYNTHETIC
		/* 0x2000 */	"annotation",		// ACC_ANNOTATION
		/* 0x4000 */	"enum"				// ACC_ENUM	
	};
	int	mask, i, seen;

	fputs(" access=", stdout);
	switch (access & 7) {
	case 0:
		fputs("default", stdout);
		break;
	case 1:
		fputs(accessName[0], stdout);
		break;
	case 2:
		fputs(accessName[1], stdout);
		break;
	case 4:
		fputs(accessName[2], stdout);
		break;
	default:
		mask = 1;
		seen = 0;
		fputc('"', stdout);
		for (i = 0; i < 3; ++i) {
			if (access & mask) {
				if (!seen) {
					seen = 1;
				} else {
					fputc(' ', stdout);
				}
				fputs(accessName[i], stdout);
			}
			mask <<= 1;
		}
		fputc('"', stdout);
	}
	access &= 0xFFF8;
	mask    = 0x8;
	seen    = 0;
	if (access) {
		for (i = 3; i < 15; ++i) {
			if (access & mask) {
				printf(" %s=y", accessName[i]);
			}
			mask <<= 1;
		}
		access &= 0x8000;
		if (access) {
			printf(" flags=0x%04x", access);
		}
	}
}

u2T *
get_datatype(u2T datatype)
{
	static u2T	classType[]  =	{'c','l','a','s','s',0};
	static u2T	byteType[]   =	{'b','y','t','e',0};
	static u2T	charType[]   =	{'c','h','a','r',0};
	static u2T	doubleType[] =	{'d','o','u','b','l','e',0};
	static u2T	floatType[]  =	{'f','l','o','a','t',0};
	static u2T	intType[]    =	{'i','n','t',0};
	static u2T	longType[]   =	{'l','o','n','g',0};
	static u2T	shortType[]  =	{'s','h','o','r','t',0};
	static u2T	booleanType[]=	{'b','o','o','l','e','a','n',0};
	static u2T	voidType[]   =	{'v','o','i','d',0};
	static u2T	unknownType[]= 	{'?',0};

	u2T *P;

	switch (datatype) {
	case 'L':
		P = classType;
		break;
	case 'B':
		P = byteType;
		break;
	case 'C':
		P = charType;
		break;
	case 'D':
		P = doubleType;
		break;
	case 'F':
		P = floatType;
		break;
	case 'I':
		P = intType;
		break;
	case 'J':
		P = longType;
		break;
	case 'S':
		P = shortType;
		break;
	case 'Z':
		P = booleanType;
		break;
	case 'V':
		P = voidType;
		break;
	default:
		P = unknownType;
	}
	return(P);
}

int
count_parameters(const u2T *descriptorP)
{
	u2T		c;
	int		seen      = 0;
	int		parameter = 0;
	
//		dumpUnicode(descriptorP);
//		fputs("->", stdout);

	for (;;) {
		switch (*descriptorP++) {
		case 0:
			return(0);
		case '[':
			continue;
		case '(':
			seen = 1;
			continue;
		case ')':
			return(parameter);
		case 'L':
			for (; (c = *descriptorP) && c != ';'; ++descriptorP);
		case 'B':
		case 'C':
		case 'D':
		case 'F':
		case 'I':
		case 'J':
		case 'S':
		case 'Z':
		case 'V':
			break;
		case ';':
			continue;
		default:
			assert(0);
		}
		if (seen) {
			++parameter;
}	}	}

void
dump_descriptor(const u2T *descriptorP)
{
	const	char *P;
	u2T	c, c1;
	int		dimension = 0;
	int		parameter = 0;
	
//		dumpUnicode(descriptorP);
//		fputs("->", stdout);

	for (;;) {
		switch (c = *descriptorP++) {
		case 0:
			return;
		case '[':
			++dimension;
			continue;
		case 'L':
			if (parameter > 1) {
				fputs(", ", stdout);
			}
			for (; (c1 = *descriptorP) && c1 != ';'; ++descriptorP) {
				if (c1 == '/') {
					c1 = '.';
				}
				fputc(c1, stdout);
			}
			goto show_array;
		case '(':
			fputc('(', stdout);
			parameter = 1;
			continue;
		case ')':
			fputs(") returns ", stdout);
			parameter = 0;
			continue;
		case 'B':
			P = "byte";
			break;
		case 'C':
			P = "char";
			break;
		case 'D':
			P = "double";
			break;
		case 'F':
			P = "float";
			break;
		case 'I':
			P = "int";
			break;
		case 'J':
			P = "long";
			break;
		case 'S':
			P = "short";
			break;
		case 'Z':
			P = "boolean";
			break;
		case 'V':
			P = "void";
			break;
		case ';':
			continue;
		default:
			assert(0);
		}
		if (parameter > 1) {
			fputs(", ", stdout);
		}
		fputs(P, stdout);
show_array:
		for (; dimension > 0; --dimension) {
			fputs("[]", stdout);
		}
		if (parameter) {
			++parameter;
}	}	}

void
deslash(u2T	*stringP)
{
	u2T	*P;

	for (P = stringP; ; ++P) {
		switch (*P) {
		case '/':
			*P = '.';
			break;
		case 0:
			return;
}	}	}


